//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// ines.cpp
//
// iNES format image handling
//
// $Id: $
//

#include <iostream>
#include <string.h>
#include "typedefs.h"
#include "cart.h"


#define  PRG_BANK_SIZE        0x4000
#define  CHR_BANK_SIZE        0x2000

#define  INES_TRAINER_LENGTH  0x0200
#define  INES_RESERVED        8


#define  VERSUS_MAPPER        99

struct iNESHeader
{
   char  magic[4];
   uint8 num_prg;
   uint8 num_chr;
   
   uint8 mirror:1;
   uint8 battery:1;
   uint8 trainer:1;
   uint8 four_screen:1;
   uint8 mapper_lonybble:4;

   uint8 mapper_hinybble;
   uint8 reserved[INES_RESERVED];
};



bool cart::isINES(const char *pFilename)
{
   FILE *fp;
   char id[5];

   fp = fopen(pFilename, "rb");
   if (NULL == fp)
      return false;

   if (fread(id, 4, 1, fp) < 1)
   {
      fclose(fp);
      return false;
   }

   fclose(fp);

   if (0 == strncmp("NES\x1A", id, 4))
      return true;

   return false;
}


int cart::loadINES()
{
   FILE *fp;
   iNESHeader header;
   char reserved[INES_RESERVED];

   fp = fopen(mpFilename, "rb");
   if (NULL == fp)
      goto _error;

   // read in the header
   if (fread(&header, sizeof(iNESHeader), 1, fp) < 1)
      goto _error;

   mMapperInfo.mNumPRG = header.num_prg;
   mMapperInfo.mNumCHR = header.num_chr;

   // iNES format doesn't tell us whether or not
   // we have SRAM, so we just allocate 8 1kB banks
   // unconditionally
   // same goes for VRAM -- just assume we've got some
   // if there isn't any CHR ROM.
   mMapperInfo.mNumSRAM = 8;
   mMapperInfo.mNumVRAM = (0 == mMapperInfo.mNumCHR) ? 1 : 0;

   mMapperInfo.mMirror = (mirror_t) header.mirror;
   
   mFlags = 0;
   if (header.battery)
      mFlags |= CART_BATTERY;
   if (header.trainer)
      mFlags |= CART_TRAINER;
   if (header.four_screen)
      mMapperInfo.mMirror = MIRROR_FOUR;

   // get original iNES 4-bit mapper number
   mMapperInfo.mMapperID = header.mapper_lonybble;

   // see if we have a clean reserved area
   memset(reserved, 0, INES_RESERVED);
   if (0 == memcmp(header.reserved, reserved, INES_RESERVED))
   {
      mMapperInfo.mMapperID |= (header.mapper_hinybble & 0xf0);
   }
   else
   {
      // silly kids, rabbits are for tricks.
      if (('D' == header.mapper_hinybble)
          && (0 == memcmp("iskDude!", header.reserved, 8)))
      {
         std::cerr << "'DiskDude!' found in header, ignoring high mapper nybble.\n";
      }
      else
      {
         std::cerr << "dirty header found, possible emulation problems.\n";
         mMapperInfo.mMapperID |= (header.mapper_hinybble & 0xf0);
      }
   }
   
   // vs unisystem
   if (VERSUS_MAPPER == mMapperInfo.mMapperID)
      mFlags |= CART_VERSUS;

   // allocate PRG/CHR memory and load them
   mMapperInfo.mpPRG = new uint8[mMapperInfo.mNumPRG * PRG_BANK_SIZE];
   if (NULL == mMapperInfo.mpPRG)
      goto _error;

   if (fread(mMapperInfo.mpPRG, PRG_BANK_SIZE, mMapperInfo.mNumPRG, fp) < mMapperInfo.mNumPRG)
      goto _error;

   if (mMapperInfo.mNumCHR)
   {
      mMapperInfo.mpCHR = new uint8[mMapperInfo.mNumCHR * CHR_BANK_SIZE];
      if (NULL == mMapperInfo.mpCHR)
         goto _error;

      if (fread(mMapperInfo.mpCHR, CHR_BANK_SIZE, mMapperInfo.mNumCHR, fp) < mMapperInfo.mNumCHR)
         goto _error;
   }

   return 0;

_error:
   if (fp)
      fclose(fp);

   return -1;
}
