//
// shatbox (C) 2002 Matthew Conte (matt@conte.com)
//
//
// msdos.cpp
//
// msdos system class
//
// $Id: $
//

#include <iostream>
#include <sys/time.h>
#include "msdos.h"
#include "event.h"
#include "config.h"


#define  US(n)    (((long long)(n)) * 1000000 / UCLOCKS_PER_SEC)

#define  DEF_VID_WIDTH        320
#define  DEF_VID_HEIGHT       240
#define  DEF_VID_BPP          32
#define  DEF_VID_SCALE        1

#define  DEF_AUD_ENABLED      1
#define  DEF_AUD_SAMPLERATE   44100

System *System_factory()
{
   return new msdos;
}


msdos::msdos()
: System()
{
   mpScreen = NULL;

   mWidth = DEF_VID_WIDTH;
   mHeight = DEF_VID_HEIGHT;
   mBPP = DEF_VID_BPP;
   mScale = DEF_VID_SCALE;

   mSoundEnabled = DEF_AUD_ENABLED;
   mSampleRate = DEF_AUD_SAMPLERATE;

   if (mpConfig)
   {
      mWidth = mpConfig->read_int("dosvideo", "width", mWidth);
      mHeight = mpConfig->read_int("dosvideo", "height", mHeight);
      mBPP = mpConfig->read_int("dosvideo", "bpp", mBPP);
      mScale = mpConfig->read_int("dosvideo", "scale", mScale);

      mSoundEnabled = mpConfig->read_int("dosaudio", "enabled", mSoundEnabled);
      mSampleRate = mpConfig->read_int("dosaudio", "samplerate", mSampleRate);
   }

   int wantmask = THIN_VIDEO | THIN_KEY;
   if (mSoundEnabled)
      wantmask |= THIN_AUDIO;

   int gotmask = thin_open(wantmask);
   if ((THIN_VIDEO | THIN_KEY) != (gotmask & (THIN_VIDEO | THIN_KEY)))
      exit(-1);

   if (0 == (gotmask & THIN_AUDIO))
      mSoundEnabled = false;

   thin_key_set_repeat(false);

   if (openVideo())
      exit(-1);

   openAudio();
}


msdos::~msdos()
{
   closeAudio();
   closeVideo();

   thin_close();
}


void msdos::openAudio()
{
   if (!mSoundEnabled)
      return;

   // set to false for now
   mSoundEnabled = false;

   thinaudio_t params;
   params.sample_rate = mSampleRate;
   params.format = THIN_AUDIO_8BIT;
   params.callback = audioCallbackDispatch;
   params.user_data = (void *) this;

   // find closest power of two samples > frag size
   int i = 1;
   while (i < mSampleRate / 60)
      i <<= 1;

   params.frag_size = i;

   if (thin_audio_open(&params))
      return;

   int buffer_size = 2 * params.frag_size;
   mpAudioStream = new unsigned char[buffer_size];
   if (NULL == mpAudioStream)
      return;

   mSoundEnabled = true;
   mAudioFragSize = buffer_size;
   mSampleRate = params.sample_rate;

   silenceAudio();
   thin_audio_pause(0);
}


void msdos::audioCallbackDispatch(void *pUserData, void *pStream, int numSamples)
{
   ((msdos *) pUserData)->audioCallback((unsigned char *) pStream, numSamples);
}


void msdos::closeAudio()
{
   if (mSoundEnabled)
   {
      thin_audio_pause(1);
      delete [] mpAudioStream;
   }
}


int msdos::openVideo()
{
   mpScreen = thin_vid_setmode(mWidth, mHeight, mBPP, THIN_VIDEO_HWSURFACE);
   if (NULL == mpScreen)
   {
      std::cerr << "could not open video mode " << mWidth
         << "x" << mHeight << "x" << mBPP << "bpp" << std::endl;
      return -1;
   }

   lockVideo();

   mBPP = mpScreen->format.bits_per_pixel;
   mPixelWidth = (mBPP + 7) / 8;
   mIndexed = mpScreen->format.indexed;

   mLoss[0] = mpScreen->format.r_loss;
   mPos[0] = mpScreen->format.r_pos;
   mLoss[1] = mpScreen->format.g_loss;
   mPos[1] = mpScreen->format.g_pos;
   mLoss[2] = mpScreen->format.b_loss;
   mPos[2] = mpScreen->format.b_pos;

   unlockVideo();

   return 0;
}


void msdos::closeVideo()
{
   thin_vid_close();
}


void msdos::setColor(int index, int r, int g, int b, int value)
{
   System::setColor(index, r, g, b, value);

   if (-1 == value && mIndexed)
   {
      color_t color;

      color.r = r;
      color.g = g;
      color.b = b;

      thin_vid_setpalette(&color, index, 1);
   }
}


void msdos::lockVideo()
{
   System::lockVideo();

   if (thin_vid_lock(mpScreen) < 0)
   {
      std::cerr << "FATAL- could not lock video surface" << std::endl;
      exit(-1);
   }

   mWidth = mpScreen->width;
   mHeight = mpScreen->height;
   mPitch = mpScreen->pitch;
   mpPixels = mpScreen->data;
}


void msdos::unlockVideo()
{
   System::unlockVideo();

   thin_vid_unlock(mpScreen);

   if (0 == mVideoLock)
      thin_vid_update(mpScreen);
}


int msdos::pollEvents(eventHandler *pEventHandler)
{
   thin_event_t thin_ev;
   thin_event_gather();

   while (thin_event_get(&thin_ev))
   {
      event ev;

      switch (thin_ev.type)
      {
      case THIN_EVENT_KEY_PRESS:
         ev.type = EVENT_NONE;

         // TODO: hacktastic for now
         switch (thin_ev.data.keysym)
         {
         case THIN_KEY_ESC:
            ev.type = EVENT_QUIT;
            break;

         case THIN_KEY_NUMPAD_PLUS:
            ev.type = EVENT_UNTHROTTLE;
            ev.data = 1;
            break;

         case THIN_KEY_R: // reset
            ev.type = EVENT_RESET;
            ev.data = 0;
            break;

         case THIN_KEY_H: // hard reset
            ev.type = EVENT_RESET;
            ev.data = 1;
            break;

         case THIN_KEY_ENTER:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_START;
            break;

         case THIN_KEY_TAB:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_SELECT;
            break;

         case THIN_KEY_LEFT_CTRL:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_B;
            break;

         case THIN_KEY_LEFT_ALT:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_A;
            break;

         case THIN_KEY_UP:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_UP;
            break;

         case THIN_KEY_DOWN:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_DOWN;
            break;

         case THIN_KEY_LEFT:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_LEFT;
            break;

         case THIN_KEY_RIGHT:
            ev.type = EVENT_PAD0_PRESS;
            ev.data = PAD_RIGHT;
            break;

         default:
            break;
         }
         pEventHandler->post(&ev);
         break;

      case THIN_EVENT_KEY_RELEASE:
         // TODO: hacktastic for now
         ev.type = EVENT_NONE;

         switch (thin_ev.data.keysym)
         {
         case THIN_KEY_NUMPAD_PLUS:
            ev.type = EVENT_UNTHROTTLE;
            ev.data = 0;
            break;

         case THIN_KEY_ENTER:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_START;
            break;

         case THIN_KEY_TAB:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_SELECT;
            break;

         case THIN_KEY_LEFT_CTRL:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_B;
            break;

         case THIN_KEY_LEFT_ALT:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_A;
            break;

         case THIN_KEY_UP:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_UP;
            break;

         case THIN_KEY_DOWN:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_DOWN;
            break;

         case THIN_KEY_LEFT:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_LEFT;
            break;

         case THIN_KEY_RIGHT:
            ev.type = EVENT_PAD0_RELEASE;
            ev.data = PAD_RIGHT;
            break;

         default:
            break;
         }
         pEventHandler->post(&ev);
         break;

      default:
         break;
      }
   }

   return 0;
}


void msdos::mark()
{
   mTimer = (long long) uclock();
}


int msdos::elapsed()
{
   uclock_t now;
   int usecs;

   now = uclock();
   usecs = US(now - mTimer);
   mTimer = now;
   
   return usecs;
}


void msdos::usleep(int us)
{
   if (us > 0)
   {
      uclock_t start = uclock();
      while (US(uclock() - start) < us)
         ;//thin_yield();
   }
}


// Reduce size of djgpp executable
extern "C" char **__crt0_glob_function(char */*_argument*/) 
{
   return (char **) 0;
}

// Reduce size of djgpp executable
extern "C" void __crt0_load_environment_file(char */*_app_name*/)
{
}
